//Standard pidgin plugin initialization
#define PURPLE_PLUGINS
#include <glib.h>
#ifndef G_GNUC_NULL_TERMINATED
# if __GNUC__ >= 4
#  define G_GNUC_NULL_TERMINATED __attribute__((__sentinel__))
# else
#  define G_GNUC_NULL_TERMINATED
# endif
#endif
#include <notify.h>
#include <plugin.h>
#include <version.h>
#include <debug.h> //Debug messages

//Configuration
const unsigned int HotKeys[]={'A','B','D','P','T'}; //The 5 hot keys (these are their virtual keycodes) [Away, Available, Offline, Show Buddy List, Toggle Hide From Taskbar]

//Includes needed for pidgin actions
#include <savedstatuses.h> //Manipulate user status
#include <status.h> //User status constants
#include <gtkblist.h> //Get the window for manipulation (pidgin_blist_get_default_gtk_blist)

//Windows stuff
#include <gdkwin32.h> //Get the window for manipulation (gdk_win32_drawable_get_handle)
static HWND DummyWindow=NULL; //This is a dummy window used to receive accelerators and hide

//Status actions
static void SetStatus(PurpleStatusPrimitive NewStatus, const char* Message, const char* StateName) //Set the state and message for the user's status
{
	purple_debug(PURPLE_DEBUG_INFO, "Status Hot Keys", "Setting Status: %s\n", StateName);
	PurpleSavedStatus *Status = purple_savedstatus_new(NULL, NewStatus); //Prepare the statu's state
	purple_savedstatus_set_message(Status, Message); //Set the status' message
	purple_savedstatus_activate(Status); //Activate the status
}
static void plugin_action_away(PurplePluginAction * action) { SetStatus(PURPLE_STATUS_AWAY, "afk or unavailable", "Away"); } //Set to away
static void plugin_action_available(PurplePluginAction * action) { SetStatus(PURPLE_STATUS_AVAILABLE, "", "Available"); } //Set to online
static void plugin_action_offline(PurplePluginAction * action) { SetStatus(PURPLE_STATUS_OFFLINE, "", "Offline"); } //Set to offline

//Show the buddy list window
static void plugin_action_show_buddy_list(PurplePluginAction * action)
{
	purple_debug(PURPLE_DEBUG_INFO, "Status Hot Keys", "Showing Buddy List\n");
	purple_blist_show();
}

//Get the buddy list window's handle
static HWND GetBuddyListWindow()
{
	GtkWidget *TheWindow=pidgin_blist_get_default_gtk_blist()->window;
	return !TheWindow ? NULL : gdk_win32_drawable_get_handle(TheWindow->window); //If window does not yet exist, return NULL, otherwise, Get the buddy list's HWND
}

//Do not show buddy list in task bar
static void plugin_action_toggle_buddy_list_in_taskbar(PurplePluginAction * action)
{
	purple_debug(PURPLE_DEBUG_INFO, "Status Hot Keys", "Swapping Buddy List taskbar toggle\n");

	//Get the window, and exit if it has not been created
	HWND BuddyListHwnd=GetBuddyListWindow();
	if(!BuddyListHwnd)
		return;

	//Toggle displaying the buddy list in the taskbar
	BOOL IsCurrentlyHidden=((HWND)GetWindowLong(BuddyListHwnd, GWL_HWNDPARENT)==DummyWindow); //Determine current taskbar status
	SetWindowLong(BuddyListHwnd, GWL_HWNDPARENT, IsCurrentlyHidden ? 0 : (LONG)DummyWindow); //Using GWL_HWNDPARENT is technically not safe in SetWindowLong, but it is the only way to do it with proper results right now

	//Force window reactiviation to toggle the taskbar
	SetForegroundWindow(DummyWindow);
	SetForegroundWindow(BuddyListHwnd);
}
static void gtkblist_created_cb(PurpleBuddyList *blist) //Do some stuff when buddy list window is created
{
	SetClassLong(DummyWindow, GCL_HICON, GetClassLong(GetBuddyListWindow(), GCL_HICON)); //Copy the buddy list icon to our dummy window so it shows when alt tabbing
	plugin_action_toggle_buddy_list_in_taskbar(NULL); //Automatically hide the buddy list in the taskbar upon buddy list window creation
}

//List of actions
static GList * plugin_actions(PurplePlugin * plugin, gpointer context)
{
	GList *list=g_list_append(NULL, purple_plugin_action_new("Away (Ctrl+Shift+A)", plugin_action_away));
	list=g_list_append(list, purple_plugin_action_new("Back [Available] (Ctrl+Shift+B)", plugin_action_available));
	list=g_list_append(list, purple_plugin_action_new("Disconnect [Offline] (Ctrl+Shift+D)", plugin_action_offline));
	list=g_list_append(list, purple_plugin_action_new("Show Pidgin Buddy List (Ctrl+Shift+P)", plugin_action_show_buddy_list));
	list=g_list_append(list, purple_plugin_action_new("Toggle Buddy List in Tashbar (Ctrl+Shift+T)", plugin_action_toggle_buddy_list_in_taskbar));
	return list;
}

//Process the dummy window's messages to find global hotkeys
static LRESULT CALLBACK WndProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	if(uMsg!=WM_HOTKEY) //Nothing to do on this window besides look for a global hotkey
		return DefWindowProc(hwnd, uMsg, wParam, lParam);

	purple_debug(PURPLE_DEBUG_INFO, "Status Hot Keys", "Processing hot key\n");

	//Determine which HotKey action was triggered
	WORD uVirtKey=(UINT) HIWORD(lParam); //Get the hotkey pressed
	int i=0;
	for(i=0;i<sizeof(HotKeys)/sizeof(UINT);i++) //Test each HotKey action against the triggered key
		if(uVirtKey==HotKeys[i])
			break;
	switch(i) //Execute the action on the triggered key
	{
		case 0: plugin_action_away(NULL); break;
		case 1: plugin_action_available(NULL); break;
		case 2: plugin_action_offline(NULL); break;
		case 3: plugin_action_show_buddy_list(NULL); break;
		case 4: plugin_action_toggle_buddy_list_in_taskbar(NULL); break;
		default: purple_debug(PURPLE_DEBUG_INFO, "Status Hot Keys", "Invalid hot key found\n");
	}

	return DefWindowProc(hwnd, uMsg, wParam, lParam);
}

//Load and unload the plugin
static gboolean plugin_load(PurplePlugin *plugin)
{
	purple_debug(PURPLE_DEBUG_INFO, "Status Hot Keys", "Loading\n");

	if(DummyWindow) //Only process shortcut keys if dummy window does not exist
		return TRUE;

	//Create a dummy window to receive the hot keys
	WNDCLASS wc={0, WndProc, 0L, 0L, GetModuleHandle(NULL), NULL, NULL, NULL, NULL, "PidginHotKeyDummy"};
	RegisterClass(&wc);
	DummyWindow=CreateWindow("PidginHotKeyDummy", "", 0, 0, 0, 0, 0, NULL, NULL, GetModuleHandle(NULL), 0);

	//Set the global hot keys
	char AtomName[]="StatusHotKeys0";
	int i;
	for(i=0;i<sizeof(HotKeys)/sizeof(UINT);i++)
	{
		AtomName[sizeof(AtomName)-1]='0'+i; //Make the atom name StatusHotKeys# where # is the key number
		RegisterHotKey(DummyWindow, GlobalAddAtom(AtomName), MOD_CONTROL|MOD_SHIFT, HotKeys[i]); //Always use ctrl+shift
	}

	//Do some stuff when buddy list window is created
	if(GetBuddyListWindow()) //If window is already created, immediately run
		gtkblist_created_cb(NULL);
	else //Otherwise, wait for the signal that it has loaded
		purple_signal_connect(pidgin_blist_get_handle(), "gtkblist-created", plugin, PURPLE_CALLBACK(gtkblist_created_cb), NULL);

	return TRUE;
}

static gboolean plugin_unload(PurplePlugin *plugin)
{
	purple_debug(PURPLE_DEBUG_INFO, "Status Hot Keys", "Unloading\n");
	purple_signals_disconnect_by_handle(plugin); //Unhook signals

	if(!DummyWindow) //Only remove shortcut keys if dummy window exists
		return TRUE;

	//Unhook the global hot keys
	char AtomName[]="StatusHotKeys0";
	int i;
	for(i=0;i<sizeof(HotKeys)/sizeof(UINT);i++)
	{
		AtomName[sizeof(AtomName)-1]='0'+i; //Make the atom name StatusHotKeys# where # is the key number
		ATOM HotKeyAtom=GlobalFindAtom(AtomName); //Find the atom
		UnregisterHotKey(DummyWindow, HotKeyAtom); //Remove the hotkey
		GlobalDeleteAtom(HotKeyAtom); //Delete the atom
	}

	//Make sure the buddy list isn't a child of the dummy window
	HWND BuddyListHwnd=GetBuddyListWindow();
	if(BuddyListHwnd && (HWND)GetWindowLong(BuddyListHwnd, GWL_HWNDPARENT)==DummyWindow) //If it is...
		plugin_action_toggle_buddy_list_in_taskbar(NULL); //Make it its own parent again

	//Destroy the dummy window
	DestroyWindow(DummyWindow);
	DummyWindow=NULL;

	return TRUE;
}

//Plugin base info
static PurplePluginInfo info={PURPLE_PLUGIN_MAGIC, PURPLE_MAJOR_VERSION, PURPLE_MINOR_VERSION, PURPLE_PLUGIN_STANDARD, NULL, 0, NULL, PURPLE_PRIORITY_DEFAULT,
	"core-StatusHotKeys", "Status Hot Keys", "1.0",
	"Global shortcut keys to set status and manipulate the buddy window visibility",
	"Global shortcut keys to: Set status to away, available, or offline; Move pidgin Buddy List to top of windows; Hides the Buddy List from the taskbar.",
	"Jeffrey Riaboy (http://www.castledragmire.com/Contact)",
	"http://www.castledragmire.com/Posts/Pidgin_Set_Status_Plugin",
	plugin_load, plugin_unload, NULL, NULL, NULL, NULL, plugin_actions, NULL, NULL, NULL, NULL};

static void init_plugin(PurplePlugin * plugin) {}
PURPLE_INIT_PLUGIN(StatusHotKeys, init_plugin, info)
